@@ -3,29 +3,33 @@
   dir.c -
 
   $Author: matz $
-  $Date: 1996/12/25 10:42:28 $
+  $Date: 1998/11/09 09:11:48 $
   created at: Wed Jan  5 09:51:01 JST 1994
 
-  Copyright (C) 1993-1996 Yukihiro Matsumoto
+  Copyright (C) 1993-1998 Yukihiro Matsumoto
 
 ************************************************/
 
 #include "ruby.h"
 
 #include <sys/types.h>
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#endif
+#include <sys/stat.h>
 
 #ifdef HAVE_SYS_PARAM_H
 # include <sys/param.h>
 #else
 # define MAXPATHLEN 1024
 #endif
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
 
 #if HAVE_DIRENT_H
 # include <dirent.h>
 # define NAMLEN(dirent) strlen((dirent)->d_name)
+#elif HAVE_DIRECT_H
+# include <direct.h>
+# define NAMLEN(dirent) strlen((dirent)->d_name)
 #else
 # define dirent direct
 # define NAMLEN(dirent) (dirent)->d_namlen
@@ -38,7 +42,7 @@
 # if HAVE_NDIR_H
 #  include <ndir.h>
 # endif
-# ifdef NT
+# if defined(NT) && defined(_MSC_VER)
 #  include "missing/dir.h"
 # endif
 #endif
@@ -49,6 +53,10 @@
 char *getenv();
 #endif
 
+#ifdef USE_CWGUSI
+# include <sys/errno.h>
+#endif
+
 static VALUE cDir;
 
 static void
@@ -58,29 +66,34 @@ free_dir(dir)
     if (dir) closedir(dir);
 }
 
+static VALUE dir_close _((VALUE));
+
 static VALUE
 dir_s_open(dir_class, dirname)
-    VALUE dir_class;
-    struct RString *dirname;
+    VALUE dir_class, dirname;
 {
     VALUE obj;
     DIR *dirp;
 
     Check_SafeStr(dirname);
 
-    dirp = opendir(dirname->ptr);
+    dirp = opendir(RSTRING(dirname)->ptr);
     if (dirp == NULL) {
 	if (errno == EMFILE || errno == ENFILE) {
-	    gc();
-	    dirp = opendir(dirname->ptr);
+	    gc_gc();
+	    dirp = opendir(RSTRING(dirname)->ptr);
 	}
 	if (dirp == NULL) {
-	    rb_sys_fail(dirname->ptr);
+	    rb_sys_fail(RSTRING(dirname)->ptr);
 	}
     }
 
     obj = Data_Wrap_Struct(dir_class, 0, free_dir, dirp);
 
+    if (iterator_p()) {
+	return rb_ensure(rb_yield, obj, dir_close, obj);
+    }
+
     return obj;
 }
 
@@ -95,6 +108,26 @@ dir_closed()
     if (dirp == NULL) dir_closed();\
 }
 
+static VALUE
+dir_read(dir)
+    VALUE dir;
+{
+    DIR *dirp;
+    struct dirent *dp;
+
+    GetDIR(dir, dirp);
+    errno = 0;
+    dp = readdir(dirp);
+    if (dp)
+	return str_taint(str_new(dp->d_name, NAMLEN(dp)));
+    else if (errno == 0) {	/* end of stream */
+	return Qnil;
+    }
+    else {
+	rb_sys_fail(0);
+    }
+}
+
 static VALUE
 dir_each(dir)
     VALUE dir;
@@ -118,7 +151,7 @@ dir_tell(dir)
     DIR *dirp;
     int pos;
 
-#if !defined(__CYGWIN32__)
+#if !defined(__CYGWIN32__) && !defined(__BEOS__)
     GetDIR(dir, dirp);
     pos = telldir(dirp);
     return int2inum(pos);
@@ -133,7 +166,7 @@ dir_seek(dir, pos)
 {
     DIR *dirp;
 
-#if !defined(__CYGWIN32__)
+#if !defined(__CYGWIN32__) && !defined(__BEOS__)
     GetDIR(dir, dirp);
     seekdir(dirp, NUM2INT(pos));
     return dir;
@@ -177,8 +210,7 @@ dir_s_chdir(argc, argv, obj)
     char *dist = "";
 
     rb_secure(2);
-    rb_scan_args(argc, argv, "01", &path);
-    if (path) {
+    if (rb_scan_args(argc, argv, "01", &path) == 1) {
 	Check_SafeStr(path);
 	dist = RSTRING(path)->ptr;
     }
@@ -190,7 +222,7 @@ dir_s_chdir(argc, argv, obj)
     }
 
     if (chdir(dist) < 0)
-	rb_sys_fail(0);
+	rb_sys_fail(dist);
 
     return INT2FIX(0);
 }
@@ -199,12 +231,12 @@ static VALUE
 dir_s_getwd(dir)
     VALUE dir;
 {
-    extern char *getwd();
     char path[MAXPATHLEN];
 
 #ifdef HAVE_GETCWD
     if (getcwd(path, sizeof(path)) == 0) rb_sys_fail(path);
 #else
+    extern char *getwd();
     if (getwd(path) == 0) rb_sys_fail(path);
 #endif
 
@@ -215,12 +247,12 @@ static VALUE
 dir_s_chroot(dir, path)
     VALUE dir, path;
 {
-#if !defined(DJGPP) && !defined(__CYGWIN32__)  && !defined(NT) && !defined(__human68k__)
+#if !defined(DJGPP) && !defined(NT) && !defined(__human68k__) && !defined(USE_CWGUSI) && !defined(__BEOS__)
     rb_secure(2);
     Check_SafeStr(path);
 
     if (chroot(RSTRING(path)->ptr) == -1)
-	rb_sys_fail(0);
+	rb_sys_fail(RSTRING(path)->ptr);
 
     return INT2FIX(0);
 #else
@@ -246,7 +278,7 @@ dir_s_mkdir(argc, argv, obj)
     }
 
     Check_SafeStr(path);
-#ifndef NT
+#if !defined(NT) && !defined(USE_CWGUSI)
     if (mkdir(RSTRING(path)->ptr, mode) == -1)
 	rb_sys_fail(RSTRING(path)->ptr);
 #else
@@ -259,13 +291,12 @@ dir_s_mkdir(argc, argv, obj)
 
 static VALUE
 dir_s_rmdir(obj, dir)
-    VALUE obj;
-    struct RString *dir;
+    VALUE obj, dir;
 {
     rb_secure(2);
     Check_SafeStr(dir);
-    if (rmdir(dir->ptr) < 0)
-	rb_sys_fail(dir->ptr);
+    if (rmdir(RSTRING(dir)->ptr) < 0)
+	rb_sys_fail(RSTRING(dir)->ptr);
 
     return TRUE;
 }
@@ -343,8 +374,7 @@ push_braces(ary, s)
 
 static VALUE
 dir_s_glob(dir, str)
-    VALUE dir;
-    struct RString *str;
+    VALUE dir, str;
 {
     char *p, *pend;
     char buf[MAXPATHLEN];
@@ -353,11 +383,13 @@ dir_s_glob(dir, str)
     VALUE ary;
 
     Check_SafeStr(str);
-
+    if (RSTRING(str)->len > MAXPATHLEN) {
+	ArgError("pathname too long (%d bytes)", RSTRING(str)->len);
+    }
     ary = ary_new();
 
-    p = str->ptr;
-    pend = p + str->len;
+    p = RSTRING(str)->ptr;
+    pend = p + RSTRING(str)->len;
 
     while (p < pend) {
 	t = buf;
@@ -386,27 +418,37 @@ dir_s_glob(dir, str)
 
 static VALUE
 dir_foreach(io, dirname)
-    VALUE io;
-    struct RString *dirname;
+    VALUE io, dirname;
 {
     VALUE dir;
 
-    dir = dir_s_open(cDir, dirname);
+    dir = rb_funcall(cDir, rb_intern("open"), 1, dirname);
     return rb_ensure(dir_each, dir, dir_close, dir);
 }
 
+static VALUE
+dir_entries(io, dirname)
+    VALUE io, dirname;
+{
+    VALUE dir;
+
+    dir = rb_funcall(cDir, rb_intern("open"), 1, dirname);
+    return rb_ensure(rb_Array, dir, dir_close, dir);
+}
+
 void
 Init_Dir()
 {
-    extern VALUE mEnumerable;
-
     cDir = rb_define_class("Dir", cObject);
 
     rb_include_module(cDir, mEnumerable);
 
+    rb_define_singleton_method(cDir, "new", dir_s_open, 1);
     rb_define_singleton_method(cDir, "open", dir_s_open, 1);
     rb_define_singleton_method(cDir, "foreach", dir_foreach, 1);
+    rb_define_singleton_method(cDir, "entries", dir_entries, 1);
 
+    rb_define_method(cDir,"read", dir_read, 0);
     rb_define_method(cDir,"each", dir_each, 0);
     rb_define_method(cDir,"rewind", dir_rewind, 0);
     rb_define_method(cDir,"tell", dir_tell, 0);
