@@ -1,20 +1,37 @@
 # jcode.rb - ruby code to handle japanese (EUC/SJIS) string
 
+$vsave, $VERBOSE = $VERBOSE, FALSE
 class String
   printf STDERR, "feel free for some warnings:\n" if $VERBOSE
 
+  def jlength
+    self.split(//).length
+  end
+
   alias original_succ succ
   private :original_succ
 
+  def mbchar?
+    if $KCODE =~ /^s/i
+      self =~ /[\x81-\x9f\xe0-\xef][\x40-\x7e\x80-\xfc]/n
+    elsif $KCODE =~ /^e/i
+      self =~ /[\xa1-\xfe][\xa1-\xfe]/n
+    else
+      false
+    end
+  end
+
   def succ
     if self[-2] && self[-2] & 0x80 != 0
       s = self.dup
       s[-1] += 1
+      s[-1] += 1 if !s.mbchar?
       return s
     else
       original_succ
     end
   end
+  alias next succ
 
   def upto(to)
     return if self > to
@@ -23,8 +40,11 @@ class String
     tail = self[-2..-1]
     if tail.length == 2 and tail  =~ /^.$/ then
       if self[0..-2] == to[0..-2]
+	first = self[-2].chr
 	for c in self[-1] .. to[-1]
-	  yield self[0..-2]+c.chr
+	  if (first+c.chr).mbchar?
+	    yield self[0..-2]+c.chr
+	  end
 	end
       end
     else
@@ -84,7 +104,7 @@ class String
   end
 
   def tr(from, to)
-    self.dup.tr!(from, to)
+    (str = self.dup).tr!(from, to) or str
   end
 
   def delete!(del)
@@ -107,7 +127,7 @@ class String
   end
 
   def delete(del)
-    self.dup.delete!(del)
+    (str = self.dup).delete!(del) or str
   end
 
   def squeeze!(del=nil)
@@ -135,7 +155,7 @@ class String
   end
 
   def squeeze(del=nil)
-    self.dup.squeeze!(del)
+    (str = self.dup).squeeze!(del) or str
   end
 
   def tr_s!(from, to)
@@ -168,7 +188,21 @@ class String
   end
 
   def tr_s(from, to)
-    self.dup.tr_s!(from,to)
+    (str = self.dup).tr_s!(from,to) or str
   end
 
+  alias original_chop! chop!
+  private :original_chop!
+
+  def chop!
+    if self =~ /(.)$/ and $1.size == 2
+      original_chop!
+    end
+    original_chop!
+  end
+
+  def chop
+    (str = self.dup).chop! or str
+  end
 end
+$VERBOSE = $vsave
