@@ -3,18 +3,17 @@
   process.c -
 
   $Author: matz $
-  $Date: 1996/12/25 10:42:47 $
+  $Date: 1998/12/14 09:00:07 $
   created at: Tue Aug 10 14:30:50 JST 1993
 
-  Copyright (C) 1993-1996 Yukihiro Matsumoto
+  Copyright (C) 1993-1998 Yukihiro Matsumoto
 
 ************************************************/
 
 #include "ruby.h"
-#include "sig.h"
+#include "rubysig.h"
 #include <stdio.h>
 #include <errno.h>
-#include <ctype.h>
 #include <signal.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -29,8 +28,9 @@ struct timeval {
 };
 #endif
 #endif /* NT */
+#include <ctype.h>
 
-struct timeval time_timeval();
+struct timeval time_timeval _((VALUE));
 
 #ifdef HAVE_SYS_WAIT_H
 # include <sys/wait.h>
@@ -43,6 +43,11 @@ struct timeval time_timeval();
 #endif
 #include "st.h"
 
+#ifdef USE_CWGUSI
+# include <sys/errno.h>
+# include "macruby_missing.h"
+#endif
+
 static VALUE
 get_pid()
 {
@@ -62,9 +67,8 @@ get_ppid()
 VALUE last_status = Qnil;
 
 #if !defined(HAVE_WAITPID) && !defined(HAVE_WAIT4)
+#define NO_WAITPID
 static st_table *pid_tbl;
-#else
-# define WAIT_CALL
 #endif
 
 static int
@@ -74,38 +78,20 @@ rb_waitpid(pid, flags, st)
     int *st;
 {
     int result;
-#if defined(THREAD) && (defined(HAVE_WAITPID) || defined(HAVE_WAIT4))
+#ifndef NO_WAITPID
+#if defined(THREAD)
     int oflags = flags;
     if (!thread_alone()) {	/* there're other threads to run */
 	flags |= WNOHANG;
     }
 #endif
 
+  retry:
 #ifdef HAVE_WAITPID
-  retry:
     result = waitpid(pid, st, flags);
-    if (result < 0) {
-	if (errno == EINTR) {
-#ifdef THREAD
-	    thread_schedule();
-#endif
-	    goto retry;
-	}
-	return -1;
-    }
-#ifdef THREAD
-    if (result == 0) {
-	if (oflags & WNOHANG) return 0;
-	thread_schedule();
-	if (thread_alone()) flags = oflags;
-	goto retry;
-    }
-#endif
-#else
-#ifdef HAVE_WAIT4
-  retry:
-
+#else  /* HAVE_WAIT4 */
     result = wait4(pid, st, flags, NULL);
+#endif
     if (result < 0) {
 	if (errno == EINTR) {
 #ifdef THREAD
@@ -123,7 +109,7 @@ rb_waitpid(pid, flags, st)
 	goto retry;
     }
 #endif
-#else
+#else  /* NO_WAITPID */
     if (pid_tbl && st_lookup(pid_tbl, pid, st)) {
 	last_status = INT2FIX(*st);
 	st_delete(pid_tbl, &pid, NULL);
@@ -151,14 +137,16 @@ rb_waitpid(pid, flags, st)
 	if (!pid_tbl)
 	    pid_tbl = st_init_numtable();
 	st_insert(pid_tbl, pid, st);
-    }
+#ifdef THREAD
+	if (!thread_alone()) thread_schedule();
 #endif
+    }
 #endif
     last_status = INT2FIX(*st);
     return result;
 }
 
-#ifndef WAIT_CALL
+#ifdef NO_WAITPID
 struct wait_data {
     int pid;
     int status;
@@ -181,7 +169,7 @@ static VALUE
 f_wait()
 {
     int pid, state;
-#ifndef WAIT_CALL
+#ifdef NO_WAITPID
     struct wait_data data;
 
     data.status = -1;
@@ -190,19 +178,21 @@ f_wait()
 	last_status = data.status;
 	return INT2FIX(data.pid);
     }
-#endif
 
     while ((pid = wait(&state)) < 0) {
-	if (errno == EINTR) {
+        if (errno == EINTR) {
 #ifdef THREAD
-	    thread_schedule();
+            thread_schedule();
 #endif
-	    continue;
-	}
-	if (errno == ECHILD) return Qnil;
-	rb_sys_fail(0);
+            continue;
+        }
+        rb_sys_fail(0);
     }
     last_status = INT2FIX(state);
+#else
+    if ((pid = rb_waitpid(-1, 0, &state)) < 0)
+	rb_sys_fail(0);
+#endif
     return INT2FIX(pid);
 }
 
@@ -223,27 +213,8 @@ f_waitpid(obj, vpid, vflags)
 char *strtok();
 
 #if defined(THREAD) && defined(HAVE_SETITIMER)
-static void
-before_exec()
-{
-    struct itimerval tval;
-
-    tval.it_interval.tv_sec = 0;
-    tval.it_interval.tv_usec = 0;
-    tval.it_value = tval.it_interval;
-    setitimer(ITIMER_VIRTUAL, &tval, NULL);
-}
-
-static void
-after_exec()
-{
-    struct itimerval tval;
-
-    tval.it_interval.tv_sec = 0;
-    tval.it_interval.tv_usec = 100000;
-    tval.it_value = tval.it_interval;
-    setitimer(ITIMER_VIRTUAL, &tval, NULL);
-}
+#define before_exec() thread_stop_timer()
+#define after_exec() thread_start_timer()
 #else
 #define before_exec()
 #define after_exec()
@@ -255,25 +226,31 @@ static void
 security(str)
     char *str;
 {
-    extern int env_path_tainted;
-    extern VALUE eSecurityError;
-
-    if (rb_safe_level() > 0 && env_path_tainted) {
-	Raise(eSecurityError, "Insecure PATH - %s", str);
+    if (rb_safe_level() > 0) {
+#ifndef USE_CWGUSI
+	if (env_path_tainted()) {
+	    Raise(eSecurityError, "Insecure PATH - %s", str);
+	}
+#endif
     }
 }
 
 static int
-proc_exec_v(argv)
+proc_exec_v(argv, prog)
     char **argv;
-{
     char *prog;
-
-    security(argv[0]);
-    prog = dln_find_exe(argv[0], 0);
-    if (!prog) {
-	errno = ENOENT;
-	return -1;
+{
+#ifndef USE_CWGUSI
+    if (prog) {
+	security(prog);
+    }
+    else {
+	security(argv[0]);
+	prog = dln_find_exe(argv[0], 0);
+	if (!prog) {
+	    errno = ENOENT;
+	    return -1;
+	}
     }
 #if (defined(MSDOS) && !defined(DJGPP)) || defined(__human68k__)
     {
@@ -312,24 +289,31 @@ proc_exec_v(argv)
     execv(prog, argv);
     after_exec();
     return -1;
+#else /* USE_CWGUSI */
+    rb_notimplement();
+#endif /* USE_CWGUSI */
 }
 
 static int
-proc_exec_n(argc, argv)
+proc_exec_n(argc, argv, progv)
     int argc;
     VALUE *argv;
+    VALUE progv;
 {
+    char *prog = 0;
     char **args;
     int i;
 
+    if (progv) {
+	prog = RSTRING(progv)->ptr;
+    }
     args = ALLOCA_N(char*, argc+1);
     for (i=0; i<argc; i++) {
-	Check_SafeStr(argv[i]);
 	args[i] = RSTRING(argv[i])->ptr;
     }
     args[i] = 0;
     if (args[0]) {
-	return proc_exec_v(args);
+	return proc_exec_v(args, prog);
     }
     return -1;
 }
@@ -338,12 +322,13 @@ int
 rb_proc_exec(str)
     char *str;
 {
+#ifndef USE_CWGUSI
     char *s = str, *t;
     char **argv, **a;
 
     security(str);
     for (s=str; *s; s++) {
-	if (*s != ' ' && !isalpha(*s) && strchr("*?{}[]<>()~&|\\$;'`\"\n",*s)) {
+	if (*s != ' ' && !ISALPHA(*s) && strchr("*?{}[]<>()~&|\\$;'`\"\n",*s)) {
 #if defined(MSDOS)
 	    int state;
 	    before_exec();
@@ -352,7 +337,7 @@ rb_proc_exec(str)
 	    if (state != -1)
 		exit(state);
 #else
-#if defined(__human68k__)
+#if defined(__human68k__) || defined(__CYGWIN32__)
 	    char *shell = dln_find_exe("sh", 0);
 	    int state = -1;
 	    before_exec();
@@ -382,24 +367,33 @@ rb_proc_exec(str)
 	*a = NULL;
     }
     if (argv[0]) {
-	return proc_exec_v(argv);
+	return proc_exec_v(argv, 0);
     }
     errno = ENOENT;
     return -1;
+#else /* USE_CWGUSI */
+    rb_notimplement();
+#endif /* USE_CWGUSI */
 }
 
 #if defined(__human68k__)
 static int
-proc_spawn_v(argv)
+proc_spawn_v(argv, prog)
     char **argv;
-{
     char *prog;
+{
     char *extension;
     int state;
 
-    prog = dln_find_exe(argv[0], 0);
-    if (!prog)
-	return -1;
+    if (prog) {
+	security(prog);
+    }
+    else {
+	security(argv[0]);
+	prog = dln_find_exe(argv[0], 0);
+	if (!prog)
+	    return -1;
+    }
 
     if ((extension = strrchr(prog, '.')) != NULL && strcasecmp(extension, ".bat") == 0) {
 	char **new_argv;
@@ -430,9 +424,10 @@ proc_spawn_v(argv)
 }
 
 static int
-proc_spawn_n(argc, argv)
+proc_spawn_n(argc, argv, prog)
     int argc;
     VALUE *argv;
+    VALUE prog;
 {
     char **args;
     int i;
@@ -442,22 +437,26 @@ proc_spawn_n(argc, argv)
 	Check_SafeStr(argv[i]);
 	args[i] = RSTRING(argv[i])->ptr;
     }
+    Check_SafeStr(prog);
     args[i] = (char *) 0;
     if (args[0])
-	return proc_exec_v(args);
+	return proc_spawn_v(args, RSTRING(prog)->ptr);
     return -1;
 }
 
 static int
-proc_spawn(str)
-    char *str;
+proc_spawn(sv)
+    VALUE sv;
 {
-    char *s = str, *t;
+    char *str;
+    char *s, *t;
     char **argv, **a;
     int state;
 
+    Check_SafeStr(sv);
+    str = s = RSTRING(sv)->ptr;
     for (s = str; *s; s++) {
-	if (*s != ' ' && !isalpha(*s) && strchr("*?{}[]<>()~&|\\$;'`\"\n",*s)) {
+	if (*s != ' ' && !ISALPHA(*s) && strchr("*?{}[]<>()~&|\\$;'`\"\n",*s)) {
 	    char *shell = dln_find_exe("sh", 0);
 	    before_exec();
 	    state = shell ? spawnl(P_WAIT, shell, "sh", "-c", str, (char *) NULL) : system(str) ;
@@ -473,7 +472,7 @@ proc_spawn(str)
 	    *a++ = t;
 	*a = NULL;
     }
-    return argv[0] ? proc_spawn_v(argv) : -1 ;
+    return argv[0] ? proc_spawn_v(argv, 0) : -1 ;
 }
 #endif /* __human68k__ */
 
@@ -482,12 +481,38 @@ f_exec(argc, argv)
     int argc;
     VALUE *argv;
 {
-    if (argc == 1) {
-	Check_SafeStr(argv[0]);
+    VALUE prog = 0;
+    int i;
+
+    if (argc == 0) {
+	ArgError("wrong # of arguments");
+    }
+    if (TYPE(argv[0]) == T_ARRAY) {
+	if (RARRAY(argv[0])->len != 2) {
+	    ArgError("wrong first argument");
+	}
+	prog = RARRAY(argv[0])->ptr[0];
+	argv[0] = RARRAY(argv[0])->ptr[1];
+    }
+
+    if (TYPE(argv[0]) == T_ARRAY) {
+	if (RARRAY(argv[0])->len != 2) {
+	    ArgError("wrong first argument");
+	}
+	prog = RARRAY(argv[0])->ptr[0];
+	argv[0] = RARRAY(argv[0])->ptr[1];
+    }
+    if (prog) {
+	Check_SafeStr(prog);
+    }
+    for (i = 0; i < argc; i++) {
+	Check_SafeStr(argv[i]);
+    }
+    if (argc == 1 && prog == 0) {
 	rb_proc_exec(RSTRING(argv[0])->ptr);
     }
     else {
-	proc_exec_n(argc, argv);
+	proc_exec_n(argc, argv, prog);
     }
     rb_sys_fail(RSTRING(argv[0])->ptr);
 }
@@ -534,7 +559,11 @@ f_exit_bang(obj, status)
 	code = INT2FIX(status);
     }
 
+#ifdef USE_CWGUSI
+    exit(code);
+#else
     _exit(code);
+#endif
 
     /* not reached */
 }
@@ -545,6 +574,7 @@ rb_syswait(pid)
 {
     RETSIGTYPE (*hfunc)(), (*qfunc)(), (*ifunc)();
     int status;
+    int i;
 
 #ifdef SIGHUP
     hfunc = signal(SIGHUP, SIG_IGN);
@@ -554,7 +584,9 @@ rb_syswait(pid)
 #endif
     ifunc = signal(SIGINT, SIG_IGN);
 
-    if (rb_waitpid(pid, 0, &status) < 0) rb_sys_fail("wait");
+    do {
+	i = rb_waitpid(pid, 0, &status);
+    } while (i == -1 && errno == EINTR);
 
 #ifdef SIGHUP
     signal(SIGHUP, hfunc);
@@ -574,6 +606,12 @@ f_system(argc, argv)
     VALUE cmd;
     int state;
 
+    if (TYPE(argv[0]) == T_ARRAY) {
+	if (RARRAY(argv[0])->len != 2) {
+	    ArgError("wrong first argument");
+	}
+	argv[0] = RARRAY(argv[0])->ptr[0];
+    }
     cmd = ary_join(ary_new4(argc, argv), str_new2(" "));
 
     Check_SafeStr(cmd);
@@ -587,6 +625,12 @@ f_system(argc, argv)
     VALUE cmd;
     int state;
 
+    if (TYPE(argv[0]) == T_ARRAY) {
+	if (RARRAY(argv[0])->len != 2) {
+	    ArgError("wrong first argument");
+	}
+	argv[0] = RARRAY(argv[0])->ptr[0];
+    }
     cmd = ary_join(ary_new4(argc, argv), str_new2(" "));
 
     Check_SafeStr(cmd);
@@ -597,6 +641,7 @@ f_system(argc, argv)
     return FALSE;
 #else
 #if defined(__human68k__)
+    VALUE prog = 0;
     int i;
     int state;
 
@@ -608,18 +653,27 @@ f_system(argc, argv)
 	return INT2FIX(0);
     }
 
-    for (i = 0; i < argc; i++)
-	Check_SafeStr(argv[i]);
+    if (TYPE(argv[0]) == T_ARRAY) {
+	if (RARRAY(argv[0])->len != 2) {
+	    ArgError("wrong first argument");
+	}
+	prog = RARRAY(argv[0])->ptr[0];
+	argv[0] = RARRAY(argv[0])->ptr[1];
+    }
 
-    state = argc == 1 ? proc_spawn(RSTRING(argv[0])->ptr) : proc_spawn_n(argc, argv) ;
+    if (argc == 1 && prog == 0) {
+	state = proc_spawn(argv[0]);
+    }
+    else {
+	state = proc_spawn_n(argc, argv, prog);
+    }
     last_status = state == -1 ? INT2FIX(127) : INT2FIX(state);
-
     return state == 0 ? TRUE : FALSE ;
 #else
-    int i;
+    volatile VALUE prog = 0;
     int pid;
+    int i;
 
-    fflush(stdin);		/* is it really needed? */
     fflush(stdout);
     fflush(stderr);
     if (argc == 0) {
@@ -627,18 +681,28 @@ f_system(argc, argv)
 	return INT2FIX(0);
     }
 
-    for (i=0; i<argc; i++) {
-	Check_SafeStr(argv[i]);
+    if (TYPE(argv[0]) == T_ARRAY) {
+	if (RARRAY(argv[0])->len != 2) {
+	    ArgError("wrong first argument");
+	}
+	prog = RARRAY(argv[0])->ptr[0];
+	argv[0] = RARRAY(argv[0])->ptr[1];
     }
 
+    if (prog) {
+	Check_SafeStr(prog);
+    }
+    for (i = 0; i < argc; i++) {
+	Check_SafeStr(argv[i]);
+    }
   retry:
     switch (pid = vfork()) {
       case 0:
-	if (argc == 1) {
+	if (argc == 1 && prog == 0) {
 	    rb_proc_exec(RSTRING(argv[0])->ptr);
 	}
 	else {
-	    proc_exec_n(argc, argv);
+	    proc_exec_n(argc, argv, prog);
 	}
 	_exit(127);
 	break;			/* not reached */
@@ -707,7 +771,7 @@ f_sleep(argc, argv)
     return INT2FIX(end);
 }
 
-#if !defined(NT) && !defined(DJGPP) && !defined(__human68k__)
+#if !defined(NT) && !defined(DJGPP) && !defined(__human68k__) && !defined(USE_CWGUSI)
 static VALUE
 proc_getpgrp(argc, argv)
     int argc;
@@ -719,7 +783,7 @@ proc_getpgrp(argc, argv)
     int pid;
 
     rb_scan_args(argc, argv, "01", &vpid);
-    pid = NUM2INT(vpid);
+    pid = NIL_P(vpid)?0:NUM2INT(vpid);
     pgrp = BSD_GETPGRP(pid);
 #else
     rb_scan_args(argc, argv, "0");
@@ -734,27 +798,31 @@ proc_setpgrp(argc, argv)
     int argc;
     VALUE *argv;
 {
+#ifdef HAVE_SETPGRP
 #ifdef BSD_SETPGRP
     VALUE pid, pgrp;
     int ipid, ipgrp;
 
     rb_scan_args(argc, argv, "02", &pid, &pgrp);
 
-    ipid = NUM2INT(pid);
-    ipgrp = NUM2INT(pgrp);
+    ipid = NIL_P(pid)?0:NUM2INT(pid);
+    ipgrp = NIL_P(pgrp)?0:NUM2INT(pgrp);
     if (BSD_SETPGRP(ipid, ipgrp) < 0) rb_sys_fail(0);
 #else
     rb_scan_args(argc, argv, "0");
     if (setpgrp() < 0) rb_sys_fail(0);
 #endif
     return Qnil;
+#else
+    rb_notimplement();
+#endif
 }
 
-#ifdef HAVE_SETPGID
 static VALUE
 proc_setpgid(obj, pid, pgrp)
     VALUE obj, pid, pgrp;
 {
+#ifdef HAVE_SETPGID
     int ipid, ipgrp;
 
     ipid = NUM2INT(pid);
@@ -762,8 +830,23 @@ proc_setpgid(obj, pid, pgrp)
 
     if (setpgid(ipid, ipgrp) < 0) rb_sys_fail(0);
     return Qnil;
-}
+#else
+    rb_notimplement();
 #endif
+}
+
+static VALUE
+proc_setsid()
+{
+#ifdef HAVE_SETSID
+    int pid = setsid();
+
+    if (pid < 0) rb_sys_fail(0);
+    return INT2FIX(pid);
+#else
+    rb_notimplement();
+#endif
+}
 
 static VALUE
 proc_getpriority(obj, which, who)
@@ -818,11 +901,11 @@ proc_setuid(obj, id)
     int uid;
 
     uid = NUM2INT(id);
-#ifdef HAVE_SETRUID
-    setruid(uid);
-#else
 #ifdef HAVE_SETREUID
     setreuid(uid, -1);
+#else
+#ifdef HAVE_SETRUID
+    setruid(uid);
 #else
     {
 	if (geteuid() == uid)
@@ -930,16 +1013,20 @@ extern VALUE f_kill();
 void
 Init_process()
 {
-    extern VALUE mKernel;
-
+#ifndef USE_CWGUSI
     rb_define_virtual_variable("$$", get_pid, 0);
+#endif
     rb_define_readonly_variable("$?", &last_status);
+#ifndef USE_CWGUSI
     rb_define_global_function("exec", f_exec, -1);
-#ifndef NT
+#endif
+#if !defined(NT) && !defined(USE_CWGUSI)
     rb_define_global_function("fork", f_fork, 0);
 #endif
     rb_define_global_function("exit!", f_exit_bang, 1);
+#ifndef USE_CWGUSI
     rb_define_global_function("system", f_system, -1);
+#endif
     rb_define_global_function("sleep", f_sleep, -1);
 
     mProcess = rb_define_module("Process");
@@ -957,30 +1044,34 @@ Init_process()
 #endif
 #endif
 
-#ifndef NT
+#if !defined(NT) && !defined(USE_CWGUSI)
     rb_define_singleton_method(mProcess, "fork", f_fork, 0);
 #endif
     rb_define_singleton_method(mProcess, "exit!", f_exit_bang, 1);
+#ifndef USE_CWGUSI
     rb_define_module_function(mProcess, "kill", f_kill, -1);
+#endif
 #ifndef NT
     rb_define_module_function(mProcess, "wait", f_wait, 0);
     rb_define_module_function(mProcess, "waitpid", f_waitpid, 2);
 
+#ifndef USE_CWGUSI
     rb_define_module_function(mProcess, "pid", get_pid, 0);
     rb_define_module_function(mProcess, "ppid", get_ppid, 0);
-#endif
+#endif /* ifndef USE_CWGUSI */
+#endif /* ifndef NT */
 
-#if !defined(NT) && !defined(DJGPP) && !defined(__human68k__)
+#if !defined(NT) && !defined(DJGPP) && !defined(__human68k__) && !defined(USE_CWGUSI)
     rb_define_module_function(mProcess, "getpgrp", proc_getpgrp, -1);
     rb_define_module_function(mProcess, "setpgrp", proc_setpgrp, -1);
-#ifdef HAVE_SETPGID
     rb_define_module_function(mProcess, "setpgid", proc_setpgid, 2);
-#endif
 
-#ifdef HAVE_GETPRIORITY
+    rb_define_module_function(mProcess, "setsid", proc_setsid, 0);
+
     rb_define_module_function(mProcess, "getpriority", proc_getpriority, 2);
     rb_define_module_function(mProcess, "setpriority", proc_setpriority, 3);
 
+#ifdef PRIO_PROCESS
     rb_define_const(mProcess, "PRIO_PROCESS", INT2FIX(PRIO_PROCESS));
     rb_define_const(mProcess, "PRIO_PGRP", INT2FIX(PRIO_PGRP));
     rb_define_const(mProcess, "PRIO_USER", INT2FIX(PRIO_USER));
