@@ -3,35 +3,56 @@
   ruby.c -
 
   $Author: matz $
-  $Date: 1996/12/25 10:42:51 $
+  $Date: 1998/12/14 09:00:07 $
   created at: Tue Aug 10 12:47:31 JST 1993
 
-  Copyright (C) 1993-1996 Yukihiro Matsumoto
+  Copyright (C) 1993-1998 Yukihiro Matsumoto
 
 ************************************************/
 
+#ifdef _WIN32
+#include <windows.h>
+#endif
 #include "ruby.h"
-#include "re.h"
 #include "dln.h"
 #include <stdio.h>
-#include <ctype.h>
 #include <sys/types.h>
 #include <fcntl.h>
+#include <ctype.h>
 
-#ifdef HAVE_STRING_H
-# include <string.h>
-#else
-char *strchr();
-char *strrchr();
-char *strstr();
+#ifdef __hpux
+#include <sys/pstat.h>
 #endif
 
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
+#ifdef __MWERKS__
+#include "node.h"
+#endif
+
+void show_version();
+void show_copyright();
+
+#ifdef USE_CWGUSI
+#include "macruby_missing.h"
+#endif
+
+#ifndef HAVE_STRING_H
+char *strchr _((char*,char));
+char *strrchr _((char*,char));
+char *strstr _((char*,char*));
+#endif
+
+char *ruby_mktemp _((void));
+
 char *getenv();
 
 static int version, copyright;
 
-int debug = FALSE;
-int verbose = FALSE;
+VALUE debug = FALSE;
+VALUE verbose = FALSE;
 int tainting = FALSE;
 static int sflag = FALSE;
 
@@ -45,34 +66,52 @@ extern int nerrs;
 static int xflag = FALSE;
 extern VALUE RS, RS_default, ORS, FS;
 
-static void load_stdin();
+static void load_stdin _((void));
 static void load_file _((char *, int));
 static void forbid_setid _((char *));
 
-static int do_loop = FALSE, do_print = FALSE;
-static int do_check = FALSE, do_line = FALSE;
-static int do_split = FALSE;
+static VALUE do_loop = FALSE, do_print = FALSE;
+static VALUE do_check = FALSE, do_line = FALSE;
+static VALUE do_split = FALSE;
 
 static char *script;
 
+static int origargc;
+static char **origargv;
+
+extern int   sourceline;
+extern char *sourcefile;
+
 #ifndef RUBY_LIB
 #define RUBY_LIB "/usr/local/lib/ruby"
 #endif
+#ifndef RUBY_SITE_LIB
+#define RUBY_SITE_LIB "/usr/local/lib/site_ruby"
+#endif
 
-#if defined(MSDOS) || defined(NT)
+#if defined(MSDOS) || defined(NT) || defined(__MACOS__)
 #define RUBY_LIB_SEP ';'
 #else
 #define RUBY_LIB_SEP ':'
 #endif
 
 extern VALUE rb_load_path;
-VALUE Frequire();
+
+static FILE *e_fp;
+static char *e_tmpname;
 
 static void
 addpath(path)
     char *path;
 {
     if (path == 0) return;
+#if defined(__CYGWIN32__)
+    {
+	char rubylib[FILENAME_MAX];
+	conv_to_posix_path(path, rubylib);
+	path = rubylib;
+    }
+#endif
     if (strchr(path, RUBY_LIB_SEP)) {
 	char *p, *s;
 	VALUE ary = ary_new();
@@ -99,7 +138,8 @@ addpath(path)
 struct req_list {
     char *name;
     struct req_list *next;
-} *req_list;
+} req_list_head;
+struct req_list *req_list_last = &req_list_head;
 
 static void
 add_modules(mod)
@@ -109,27 +149,24 @@ add_modules(mod)
 
     list = ALLOC(struct req_list);
     list->name = mod;
-    list->next = req_list;
-    req_list = list;
+    list->next = 0;
+    req_list_last->next = list;
+    req_list_last = list;
 }
 
 void
-rb_require_modules()
+ruby_require_modules()
 {
-    struct req_list *list = req_list;
+    struct req_list *list = req_list_head.next;
     struct req_list *tmp;
-    extern void *eval_tree; /* hack to save syntax tree */
-    void *save;
 
-    req_list = 0;
-    save = eval_tree;
+    req_list_last = 0;
     while (list) {
 	f_require(Qnil, str_new2(list->name));
 	tmp = list->next;
 	free(list);
 	list = tmp;
     }
-    eval_tree = save;
 }
 
 static void
@@ -147,6 +184,7 @@ proc_options(argcp, argvp)
     version = FALSE;
     do_search = FALSE;
     script_given = 0;
+    e_tmpname = NULL;
 
     for (argc--,argv++; argc > 0; argc--,argv++) {
 	if (argv[0][0] != '-' || !argv[0][1]) break;
@@ -169,6 +207,7 @@ proc_options(argcp, argvp)
 
 	  case 'd':
 	    debug = TRUE;
+	    verbose |= 1;
 	    s++;
 	    goto reswitch;
 
@@ -210,15 +249,20 @@ proc_options(argcp, argvp)
 
 	  case 'e':
 	    forbid_setid("-e");
-	    script_given++;
-	    if (script == 0) script = "-e";
+	    if (!e_fp) {
+		e_tmpname = ruby_mktemp();
+		if (!e_tmpname) Fatal("Can't mktemp");
+		e_fp = fopen(e_tmpname, "w");
+		if (!e_fp) {
+		    Fatal("Cannot open temporary file: %s", e_tmpname);
+		}
+		if (script == 0) script = e_tmpname;
+	    }
 	    if (argv[1]) {
-		compile_string("-e", argv[1], strlen(argv[1]));
-		argc--,argv++;
-	    }
-	    else {
-		compile_string("-e", "", 0);
+		fputs(argv[1], e_fp);
+		argc--, argv++;
 	    }
+	    putc('\n', e_fp);
 	    break;
 
 	  case 'r':
@@ -331,6 +375,11 @@ proc_options(argcp, argvp)
 	    }
 	    break;
 
+	  case '*':
+	  case ' ':
+	    if (s[1] == '-') s+=2;
+	    break;
+
 	  default:
 	    Fatal("Unrecognized switch: -%s",s);
 
@@ -342,6 +391,14 @@ proc_options(argcp, argvp)
   switch_end:
     if (*argvp[0] == 0) return;
 
+    if (e_fp) {
+	if (fflush(e_fp) || ferror(e_fp) || fclose(e_fp))
+	    Fatal("Cannot write to temp file for -e");
+	e_fp = NULL;
+	argc++, argv--;
+	argv[0] = e_tmpname;
+    }
+
     if (version) {
 	show_version();
 	exit(0);
@@ -350,6 +407,7 @@ proc_options(argcp, argvp)
 	show_copyright();
     }
 
+    Init_ext();		/* should be called here for some reason :-( */
     if (script_given == FALSE) {
 	if (argc == 0) {	/* no more args */
 	    if (verbose == 3) exit(0);
@@ -366,11 +424,12 @@ proc_options(argcp, argvp)
 		if (do_search) {
 		    char *path = getenv("RUBYPATH");
 
+		    script = 0;
 		    if (path) {
-			script = dln_find_file(script, path);
+			script = dln_find_file(argv[0], path);
 		    }
 		    if (!script) {
-			script = dln_find_file(script, getenv("PATH"));
+			script = dln_find_file(argv[0], getenv("PATH"));
 		    }
 		    if (!script) script = argv[0];
 		}
@@ -402,6 +461,7 @@ proc_options(argcp, argvp)
 	    else {
 		rb_gvar_set2(argv[0], TRUE);
 	    }
+	    argv[0][0] = '-';
 	}
 	*argcp = argc; *argvp = argv;
     }
@@ -421,6 +481,13 @@ load_file(fname, script)
 	f = rb_stdin;
     }
     else {
+	FILE *fp = fopen(fname, "r");
+
+	if (fp == NULL) {
+	    LoadError("No such file to load -- %s", fname);
+	}
+	fclose(fp);
+
 	f = file_open(fname, "r");
     }
 
@@ -428,6 +495,7 @@ load_file(fname, script)
 	VALUE c;
 	VALUE line;
 	VALUE rs = RS;
+	char *p;
 
 	RS = RS_default;
 	if (xflag) {
@@ -438,7 +506,7 @@ load_file(fname, script)
 		if (RSTRING(line)->len > 2
 		    && RSTRING(line)->ptr[0] == '#'
 		    && RSTRING(line)->ptr[1] == '!') {
-		    if (strstr(RSTRING(line)->ptr, "ruby")) {
+		    if (p = strstr(RSTRING(line)->ptr, "ruby")) {
 			goto start_read;
 		    }
 		}
@@ -452,25 +520,60 @@ load_file(fname, script)
 	    line = io_gets(f);
 	    line_start++;
 
-	    if (RSTRING(line)->len > 2
-		&& RSTRING(line)->ptr[0] == '!') {
+	    if (RSTRING(line)->len > 2 && RSTRING(line)->ptr[0] == '!') {
+		if ((p = strstr(RSTRING(line)->ptr, "ruby")) == 0) {
+		    /* not ruby script, kick the program */
+		    char **argv;
+		    char *path;
+		    char *pend = RSTRING(line)->ptr + RSTRING(line)->len;
 
-		char *p;
+		    p = RSTRING(line)->ptr + 2;	/* skip `#!' */
+		    if (pend[-1] == '\n') pend--; /* chomp line */
+		    if (pend[-1] == '\r') pend--;
+		    *pend = '\0';
+		    while (p < pend && ISSPACE(*p))
+			p++;
+		    path = p;	/* interpreter path */
+		    while (p < pend && !ISSPACE(*p))
+			p++;
+		    *p++ = '\0';
+		    if (p < pend) {
+			argv = ALLOCA_N(char*, origargc+3);
+			argv[1] = p;
+			MEMCPY(argv+2, origargv+1, char*, origargc);
+		    }
+		    else {
+			argv = origargv;
+		    }
+		    argv[0] = path;
+#ifndef USE_CWGUSI
+		    execv(path, argv);
+#endif
+		    sourcefile = fname;
+		    sourceline = 1;
+		    Fatal("Can't exec %s", path);
+		}
 
 	      start_read:
-		if (p = strstr(RSTRING(line)->ptr, "ruby -")) {
+		p += 4;
+		RSTRING(line)->ptr[RSTRING(line)->len-1] = '\0';
+		if (RSTRING(line)->ptr[RSTRING(line)->len-2] == '\r')
+		    RSTRING(line)->ptr[RSTRING(line)->len-2] = '\0';
+		if (p = strstr(p, " -")) {
 		    int argc; char *argv[2]; char **argvp = argv;
-		    UCHAR *s;
+		    char *s = ++p;
 
-		    s = RSTRING(line)->ptr;
-		    while (isspace(*s++))
-			;
-		    *s = '\0';
-		    RSTRING(line)->ptr[RSTRING(line)->len-1] = '\0';
-		    if (RSTRING(line)->ptr[RSTRING(line)->len-2] == '\r')
-			RSTRING(line)->ptr[RSTRING(line)->len-2] = '\0';
-		    argc = 2; argv[0] = 0; argv[1] = p + 5;
-		    proc_options(&argc, &argvp);
+		    argc = 2; argv[0] = 0;
+		    while (*p == '-') {
+			while (*s && !ISSPACE(*s))
+			    s++;
+			*s = '\0';
+			argv[1] = p;
+			proc_options(&argc, &argvp);
+			p = ++s;
+			while (*p && ISSPACE(*p))
+			    p++;
+		    }
 		}
 	    }
 	}
@@ -480,7 +583,12 @@ load_file(fname, script)
 	RS = rs;
     }
     compile_file(fname, f, line_start);
-    if (f != rb_stdin) io_close(f);
+    if (script) {
+	rb_define_global_const("DATA", f);
+    }
+    else if (f != rb_stdin) {
+	io_close(f);
+    }
 }
 
 void
@@ -497,11 +605,9 @@ load_stdin()
     load_file("-", 1);
 }
 
-VALUE Progname;
-VALUE Argv;
-
-static int origargc;
-static char **origargv;
+VALUE rb_progname;
+VALUE rb_argv;
+VALUE rb_argv0;
 
 static void
 set_arg0(val, id)
@@ -513,7 +619,7 @@ set_arg0(val, id)
     static int len;
 
     if (origargv == 0) Fail("$0 not initialized");
-    Check_Type(val, T_STRING);
+#ifndef __hpux
     if (len == 0) {
 	s = origargv[0];
 	s += strlen(s);
@@ -524,8 +630,9 @@ set_arg0(val, id)
 	}
 	len = s - origargv[0];
     }
-    s = RSTRING(val)->ptr;
-    i = RSTRING(val)->len;
+#endif
+    s = str2cstr(val, &i);
+#ifndef __hpux
     if (i > len) {
 	memcpy(origargv[0], s, len);
 	origargv[0][len] = '\0';
@@ -537,7 +644,22 @@ set_arg0(val, id)
 	while (++i < len)
 	    *s++ = ' ';
     }
-    Progname = str_taint(str_new2(origargv[0]));
+    rb_progname = str_taint(str_new2(origargv[0]));
+#else
+    if (i >= PST_CLEN) {
+      union pstun j;
+      j.pst_command = s;
+      i = PST_CLEN;
+      RSTRING(val)->len = i;
+      *(s + i) = '\0';
+      pstat(PSTAT_SETCMD, j, PST_CLEN, 0, 0);
+    } else {
+      union pstun j;
+      j.pst_command = s;
+      pstat(PSTAT_SETCMD, j, i, 0, 0);
+    }
+    rb_progname = str_taint(str_new(s, i));
+#endif
 }
 
 void
@@ -545,7 +667,7 @@ ruby_script(name)
     char *name;
 {
     if (name) {
-	Progname = str_taint(str_new2(name));
+	rb_progname = str_taint(str_new2(name));
 	sourcefile = name;
     }
 }
@@ -576,8 +698,44 @@ forbid_setid(s)
         Fatal("No %s allowed while running setuid", s);
     if (egid != gid)
         Fatal("No %s allowed while running setgid", s);
+    if (rb_safe_level() > 0)
+        Fatal("No %s allowed in tainted mode", s);
 }
 
+#if defined(_WIN32) || defined(DJGPP)
+static char *
+ruby_libpath()
+{
+    static char libpath[FILENAME_MAX+1];
+    char *p;
+#if defined(_WIN32)
+    GetModuleFileName(NULL, libpath, sizeof libpath);
+#elif defined(DJGPP)
+    extern char *__dos_argv0;
+    strcpy(libpath, __dos_argv0);
+#endif
+    p = strrchr(libpath, '\\');
+    if (p)
+	*p = 0;
+    if (!strcasecmp(p-4, "\\bin"))
+	p -= 4;
+    strcpy(p, "\\lib");
+#if defined(__CYGWIN32__)
+    p = (char *)malloc(strlen(libpath)+10);
+    if (!p)
+	return 0;
+    cygwin32_conv_to_posix_path(libpath, p);
+    strcpy(libpath, p);
+    free(p);
+#else
+    for (p = libpath; *p; p++)
+	if (*p == '\\')
+	    *p = '/';
+#endif
+    return libpath;
+}
+#endif
+
 void
 ruby_prog_init()
 {
@@ -592,27 +750,43 @@ ruby_prog_init()
     rb_define_readonly_variable("$-l", &do_line);
 
     if (rb_safe_level() == 0) {
-	addpath(getenv("RUBYLIB"));
+	addpath(".");
     }
 
-#ifdef RUBY_THIN_ARCHLIB
-    addpath(RUBY_THIN_ARCHLIB);
+    addpath(RUBY_LIB);
+#if defined(_WIN32) || defined(DJGPP)
+    addpath(ruby_libpath());
+#endif
+#ifdef __MACOS__
+    setup_macruby_libpath();
 #endif
 
 #ifdef RUBY_ARCHLIB
     addpath(RUBY_ARCHLIB);
 #endif
-    addpath(RUBY_LIB);
+#ifdef RUBY_THIN_ARCHLIB
+    addpath(RUBY_THIN_ARCHLIB);
+#endif
+
+    addpath(RUBY_SITE_LIB);
+#ifdef RUBY_SITE_ARCHLIB
+    addpath(RUBY_SITE_ARCHLIB);
+#endif
+#ifdef RUBY_SITE_THIN_ARCHLIB
+    addpath(RUBY_SITE_THIN_ARCHLIB);
+#endif
+
     if (rb_safe_level() == 0) {
-	addpath(".");
+	addpath(getenv("RUBYLIB"));
     }
 
-    rb_define_hooked_variable("$0", &Progname, 0, set_arg0);
+    rb_define_hooked_variable("$0", &rb_progname, 0, set_arg0);
 
-    Argv = ary_new();
-    rb_define_readonly_variable("$*", &Argv);
-    rb_define_global_const("ARGV", Argv);
+    rb_argv = ary_new();
+    rb_define_readonly_variable("$*", &rb_argv);
+    rb_define_global_const("ARGV", rb_argv);
     rb_define_readonly_variable("$-a", &do_split);
+    rb_global_variable(&rb_argv0);
 
 #ifdef MSDOS
     /*
@@ -636,7 +810,7 @@ ruby_set_argv(argc, argv)
     else          dln_argv0 = argv[0];
 #endif
     for (i=0; i < argc; i++) {
-	ary_push(Argv, str_taint(str_new2(argv[i])));
+	ary_push(rb_argv, str_taint(str_new2(argv[i])));
     }
 }
 
@@ -645,11 +819,9 @@ ruby_process_options(argc, argv)
     int argc;
     char **argv;
 {
-    extern VALUE errat;
-    int i;
-
     origargc = argc; origargv = argv;
     ruby_script(argv[0]);	/* for the time being */
+    rb_argv0 = rb_progname;
 #if defined(USE_DLN_A_OUT)
     dln_argv0 = argv[0];
 #endif
@@ -667,4 +839,12 @@ ruby_process_options(argc, argv)
     if (do_loop) {
 	yywhile_loop(do_line, do_split);
     }
+    if (e_fp) {
+	fclose(e_fp);
+	e_fp = NULL;
+    }
+    if (e_tmpname) {
+	unlink(e_tmpname);
+	e_tmpname = NULL;
+    }
 }
