@@ -3,18 +3,19 @@
   time.c -
 
   $Author: matz $
-  $Date: 1996/12/25 09:30:28 $
+  $Date: 1998/11/26 08:25:46 $
   created at: Tue Dec 28 14:31:59 JST 1993
 
-  Copyright (C) 1993-1996 Yukihiro Matsumoto
+  Copyright (C) 1993-1998 Yukihiro Matsumoto
 
 ************************************************/
 
 #include "ruby.h"
 #include <sys/types.h>
 
-#ifdef HAVE_STRING_H
-# include <string.h>
+#ifdef USE_CWGUSI
+int gettimeofday(struct timeval*, struct timezone*);
+int strcasecmp(char*, char*);
 #endif
 
 #include <time.h>
@@ -38,7 +39,6 @@ static VALUE cTime;
 #if defined(HAVE_TIMES) || defined(NT)
 static VALUE S_Tms;
 #endif
-extern VALUE mComparable;
 
 struct time_object {
     struct timeval tv;
@@ -54,31 +54,32 @@ struct time_object {
 }
 
 static VALUE
-time_s_now(class)
-    VALUE class;
+time_s_now(klass)
+    VALUE klass;
 {
     VALUE obj;
     struct time_object *tobj;
 
-    obj = Data_Make_Struct(class, struct time_object, 0, 0, tobj);
+    obj = Data_Make_Struct(klass, struct time_object, 0, free, tobj);
     tobj->tm_got=0;
 
     if (gettimeofday(&(tobj->tv), 0) == -1) {
 	rb_sys_fail("gettimeofday");
     }
+    obj_call_init(obj);
 
     return obj;
 }
 
 static VALUE
-time_new_internal(class, sec, usec)
-    VALUE class;
+time_new_internal(klass, sec, usec)
+    VALUE klass;
     int sec, usec;
 {
     VALUE obj;
     struct time_object *tobj;
 
-    obj = Data_Make_Struct(class, struct time_object, 0, 0, tobj);
+    obj = Data_Make_Struct(klass, struct time_object, 0, free, tobj);
     tobj->tm_got = 0;
     tobj->tv.tv_sec = sec;
     tobj->tv.tv_usec = usec;
@@ -102,23 +103,13 @@ time_timeval(time)
 
     switch (TYPE(time)) {
       case T_FIXNUM:
-	t.tv_sec = FIX2UINT(time);
-	if (t.tv_sec < 0)
-	    ArgError("time must be positive");
+	t.tv_sec = FIX2INT(time);
 	t.tv_usec = 0;
 	break;
 
       case T_FLOAT:
-	{
-	    double seconds, microseconds;
-
-	    if (RFLOAT(time)->value < 0.0)
-		ArgError("time must be positive");
-	    seconds = floor(RFLOAT(time)->value);
-	    microseconds = (RFLOAT(time)->value - seconds) * 1000000.0;
-	    t.tv_sec = seconds;
-	    t.tv_usec = microseconds;
-	}
+	t.tv_sec = RFLOAT(time)->value;
+	t.tv_usec = (RFLOAT(time)->value - t.tv_sec) * 1000000.0;
 	break;
 
       case T_BIGNUM:
@@ -139,13 +130,13 @@ time_timeval(time)
 }
 
 static VALUE
-time_s_at(class, time)
-    VALUE class, time;
+time_s_at(klass, time)
+    VALUE klass, time;
 {
     struct timeval tv;
 
     tv = time_timeval(time);
-    return time_new_internal(class, tv.tv_sec, tv.tv_usec);
+    return time_new_internal(klass, tv.tv_sec, tv.tv_usec);
 }
 
 static char *months [12] = {
@@ -153,6 +144,17 @@ static char *months [12] = {
     "jul", "aug", "sep", "oct", "nov", "dec",
 };
 
+static int
+obj2int(obj)
+    VALUE obj;
+{
+    if (TYPE(obj) == T_STRING) {
+	obj = str2inum(RSTRING(obj)->ptr, 10);
+    }
+
+    return NUM2INT(obj);
+}
+
 static void
 time_arg(argc, argv, args)
     int argc;
@@ -162,12 +164,22 @@ time_arg(argc, argv, args)
     VALUE v[6];
     int i;
 
-    rb_scan_args(argc, argv, "15", &v[0], &v[1], &v[2], &v[3], &v[4], &v[5]);
+    if (argc == 10) {
+	v[0] = argv[5];
+	v[1] = argv[4];
+	v[2] = argv[3];
+	v[3] = argv[2];
+	v[4] = argv[1];
+	v[5] = argv[0];
+    }
+    else {
+	rb_scan_args(argc, argv, "15", &v[0],&v[1],&v[2],&v[3],&v[4],&v[5]);
+    }
 
-    args[0] = NUM2INT(v[0]);
+    args[0] = obj2int(v[0]);
     if (args[0] < 70) args[0] += 100;
     if (args[0] > 1900) args[0] -= 1900;
-    if (v[1] == Qnil) {
+    if (NIL_P(v[1])) {
 	args[1] = 0;
     }
     else if (TYPE(v[1]) == T_STRING) {
@@ -182,44 +194,46 @@ time_arg(argc, argv, args)
 	    char c = RSTRING(v[1])->ptr[0];
 
 	    if ('0' <= c && c <= '9') {
-		args[1] = NUM2INT(v[1])-1;
+		args[1] = obj2int(v[1])-1;
 	    }
 	}
     }
     else {
-	args[1] = NUM2INT(v[1]);
+	args[1] = obj2int(v[1]) - 1;
     }
-    if (v[2] == Qnil) {
+    if (NIL_P(v[2])) {
 	args[2] = 1;
     }
     else {
-	args[2] = NUM2INT(v[2]);
+	args[2] = obj2int(v[2]);
     }
     for (i=3;i<6;i++) {
-	if (v[i] == Qnil) {
+	if (NIL_P(v[i])) {
 	    args[i] = 0;
 	}
 	else {
-	    args[i] = NUM2INT(v[i]);
+	    args[i] = obj2int(v[i]);
 	}
     }
 
     /* value validation */
-    if (   args[0] < 70|| args[1] > 137
+    if (   args[0] < 70|| args[0] > 137
 	|| args[1] < 0 || args[1] > 11
 	|| args[2] < 1 || args[2] > 31
 	|| args[3] < 0 || args[3] > 23
-	|| args[4] < 0 || args[4] > 60
-	|| args[5] < 0 || args[5] > 61)
+	|| args[4] < 0 || args[4] > 59
+	|| args[5] < 0 || args[5] > 60)
 	ArgError("argument out of range");
 }
 
+static VALUE time_gmtime _((VALUE));
+static VALUE time_localtime _((VALUE));
 static VALUE
-time_gm_or_local(argc, argv, gm_or_local, class)
+time_gm_or_local(argc, argv, gm_or_local, klass)
     int argc;
     VALUE *argv;
     int gm_or_local;
-    VALUE class;
+    VALUE klass;
 {
     int args[6];
     struct timeval tv;
@@ -227,6 +241,7 @@ time_gm_or_local(argc, argv, gm_or_local, class)
     time_t guess, t;
     int diff;
     struct tm *(*fn)();
+    VALUE time;
 
     fn = (gm_or_local) ? gmtime : localtime;
     time_arg(argc, argv, args);
@@ -237,7 +252,7 @@ time_gm_or_local(argc, argv, gm_or_local, class)
     tm = (*fn)(&guess);
     if (!tm) goto error;
     t = args[0];
-    while (diff = t - tm->tm_year) {
+    while (diff = t - (tm->tm_year)) {
 	guess += diff * 364 * 24 * 3600;
 	if (guess < 0) ArgError("too far future");
 	tm = (*fn)(&guess);
@@ -254,28 +269,30 @@ time_gm_or_local(argc, argv, gm_or_local, class)
     guess += (args[4] - tm->tm_min) * 60;
     guess += args[5] - tm->tm_sec;
 
-    return time_new_internal(class, guess, 0);
+    time = time_new_internal(klass, guess, 0);
+    if (gm_or_local) return time_gmtime(time);
+    return time_localtime(time);
 
   error:
-    ArgError("gmtime error");
+    ArgError("gmtime/localtime error");
 }
 
 static VALUE
-time_s_timegm(argc, argv, class)
+time_s_timegm(argc, argv, klass)
     int argc;
     VALUE *argv;
-    VALUE class;
+    VALUE klass;
 {
-    return time_gm_or_local(argc, argv, 1, class);
+    return time_gm_or_local(argc, argv, 1, klass);
 }
 
 static VALUE
-time_s_timelocal(argc, argv, class)
+time_s_timelocal(argc, argv, klass)
     int argc;
     VALUE *argv;
-    VALUE class;
+    VALUE klass;
 {
-    return time_gm_or_local(argc, argv, 0, class);
+    return time_gm_or_local(argc, argv, 0, klass);
 }
 
 static VALUE
@@ -416,6 +433,23 @@ time_gmtime(time)
 static VALUE
 time_asctime(time)
     VALUE time;
+{
+    struct time_object *tobj;
+    char *s;
+
+    GetTimeval(time, tobj);
+    if (tobj->tm_got == 0) {
+	time_localtime(time);
+    }
+    s = asctime(&(tobj->tm));
+    if (s[24] == '\n') s[24] = '\0';
+
+    return str_new2(s);
+}
+
+static VALUE
+time_to_s(time)
+    VALUE time;
 {
     struct time_object *tobj;
     char buf[64];
@@ -437,20 +471,6 @@ time_asctime(time)
     return str_new(buf, len);
 }
 
-static VALUE
-time_coerce(time1, time2)
-    VALUE time1, time2;
-{
-    if (TYPE(time2) == T_FLOAT) {
-	double d = RFLOAT(time2)->value;
-	unsigned int i = (unsigned int) d;
-
-	return assoc_new(time_new(i, (int)((d - (double)i)*1e6)),time1);
-    }
-
-    return assoc_new(time_new(NUM2INT(time2), 0), time1);
-}
-
 static VALUE
 time_plus(time1, time2)
     VALUE time1, time2;
@@ -577,7 +597,8 @@ time_mon(time)
     if (tobj->tm_got == 0) {
 	time_localtime(time);
     }
-    return INT2FIX(tobj->tm.tm_mon);
+    Warning("Time#month now start from 1 for January");
+    return INT2FIX(tobj->tm.tm_mon+1);
 }
 
 static VALUE
@@ -590,7 +611,8 @@ time_year(time)
     if (tobj->tm_got == 0) {
 	time_localtime(time);
     }
-    return INT2FIX(tobj->tm.tm_year);
+    Warning("Time#year now returns 19xx");
+    return INT2FIX(tobj->tm.tm_year+1900);
 }
 
 static VALUE
@@ -629,7 +651,8 @@ time_isdst(time)
     if (tobj->tm_got == 0) {
 	time_localtime(time);
     }
-    return INT2FIX(tobj->tm.tm_isdst);
+    Warning("Time#isdst now returns boolean value");
+    return tobj->tm.tm_isdst?TRUE:FALSE;
 }
 
 static VALUE
@@ -654,23 +677,48 @@ time_to_a(time)
     VALUE time;
 {
     struct time_object *tobj;
-    VALUE ary;
 
     GetTimeval(time, tobj);
     if (tobj->tm_got == 0) {
 	time_localtime(time);
     }
-    ary = ary_new3(9,
-		   INT2FIX(tobj->tm.tm_sec),
-		   INT2FIX(tobj->tm.tm_min),
-		   INT2FIX(tobj->tm.tm_hour),
-		   INT2FIX(tobj->tm.tm_mday),
-		   INT2FIX(tobj->tm.tm_mon),
-		   INT2FIX(tobj->tm.tm_year),
-		   INT2FIX(tobj->tm.tm_wday),
-		   INT2FIX(tobj->tm.tm_yday),
-		   INT2FIX(tobj->tm.tm_isdst));
-    return ary;
+    Warning("Time#to_a's return values are now changed");
+    return ary_new3(10,
+		    INT2FIX(tobj->tm.tm_sec),
+		    INT2FIX(tobj->tm.tm_min),
+		    INT2FIX(tobj->tm.tm_hour),
+		    INT2FIX(tobj->tm.tm_mday),
+		    INT2FIX(tobj->tm.tm_mon+1),
+		    INT2FIX(tobj->tm.tm_year+1900),
+		    INT2FIX(tobj->tm.tm_wday),
+		    INT2FIX(tobj->tm.tm_yday),
+		    tobj->tm.tm_isdst?TRUE:FALSE,
+		    time_zone(time));
+}
+
+#define SMALLBUF 100
+static int
+rb_strftime(buf, format, time)
+    char ** volatile buf;
+    char * volatile format;
+    struct tm * volatile time;
+{
+    volatile int i;
+    int len;
+
+    len = strftime(*buf, SMALLBUF, format, time);
+    if (len != 0) return len;
+    for (i=1024; i<8192; i+=1024) {
+	*buf = xmalloc(i);
+	len = strftime(*buf, i-1, format, time);
+	if (len == 0) {
+	    free(*buf);
+	    continue;
+	}
+	return len;
+    }
+
+    ArgError("bad strftime format or result too long");
 }
 
 static VALUE
@@ -678,31 +726,34 @@ time_strftime(time, format)
     VALUE time, format;
 {
     struct time_object *tobj;
-    char buf[100];
+    char buffer[SMALLBUF];
+    char *fmt;
+    char *buf = buffer;
     int len;
+    VALUE str;
 
-    Check_Type(format, T_STRING);
     GetTimeval(time, tobj);
     if (tobj->tm_got == 0) {
 	time_localtime(time);
     }
-    if (strlen(RSTRING(format)->ptr) < RSTRING(format)->len) {
-	/* Ruby string contains \0. */
-	VALUE str;
-	int l;
-	char *p = RSTRING(format)->ptr, *pe = p + RSTRING(format)->len;
+    fmt = str2cstr(format, &len);
+    if (strlen(fmt) < len) {
+	/* Ruby string may contain \0's. */
+	char *p = fmt, *pe = fmt + len;
 
 	str = str_new(0, 0);
 	while (p < pe) {
-	    len = strftime(buf, 100, p, &(tobj->tm));
+	    len = rb_strftime(&buf, p, &(tobj->tm));
 	    str_cat(str, buf, len);
-	    l = strlen(p);
-	    p += l + 1;
+	    p += strlen(p) + 1;
+	    if (len > SMALLBUF) free(buf);
 	}
 	return str;
     }
-    len = strftime(buf, 100, RSTRING(format)->ptr, &(tobj->tm));
-    return str_new(buf, len);
+    len = rb_strftime(&buf, RSTRING(format)->ptr, &(tobj->tm));
+    str = str_new(buf, len);
+    if (buf != buffer) free(buf);
+    return str;
 }
 
 static VALUE
@@ -739,6 +790,54 @@ time_s_times(obj)
 #endif
 }
 
+static VALUE
+time_dump(time, limit)
+    VALUE time, limit;
+{
+    struct time_object *tobj;
+    int sec, usec;
+    unsigned char buf[8];
+    int i;
+
+    GetTimeval(time, tobj);
+    sec = tobj->tv.tv_sec;
+    usec = tobj->tv.tv_usec;
+
+    for (i=0; i<4; i++) {
+	buf[i] = sec & 0xff;
+	sec = RSHIFT(sec, 8);
+    }
+    for (i=4; i<8; i++) {
+	buf[i] = usec & 0xff;
+	usec = RSHIFT(usec, 8);
+    }
+    return str_new(buf, 8);
+}
+
+static VALUE
+time_load(klass, str)
+    VALUE klass, str;
+{
+    int sec, usec;
+    unsigned char *buf;
+    int i;
+
+    buf = str2cstr(str, &i);
+    if (i != 8) {
+	TypeError("marshaled time format differ");
+    }
+
+    sec = usec = 0;
+    for (i=0; i<4; i++) {
+	sec |= buf[i]<<(8*i);
+    }
+    for (i=4; i<8; i++) {
+	usec |= buf[i]<<(8*(i-4));
+    }
+
+    return time_new_internal(klass, sec, usec);
+}
+
 void
 Init_Time()
 {
@@ -764,10 +863,9 @@ Init_Time()
     rb_define_method(cTime, "gmtime", time_gmtime, 0);
     rb_define_method(cTime, "ctime", time_asctime, 0);
     rb_define_method(cTime, "asctime", time_asctime, 0);
-    rb_define_method(cTime, "to_s", time_asctime, 0);
-    rb_define_method(cTime, "inspect", time_asctime, 0);
+    rb_define_method(cTime, "to_s", time_to_s, 0);
+    rb_define_method(cTime, "inspect", time_to_s, 0);
     rb_define_method(cTime, "to_a", time_to_a, 0);
-    rb_define_method(cTime, "coerce", time_coerce, 1);
 
     rb_define_method(cTime, "+", time_plus, 1);
     rb_define_method(cTime, "-", time_minus, 1);
@@ -776,7 +874,9 @@ Init_Time()
     rb_define_method(cTime, "min", time_min, 0);
     rb_define_method(cTime, "hour", time_hour, 0);
     rb_define_method(cTime, "mday", time_mday, 0);
+    rb_define_method(cTime, "day", time_mday, 0);
     rb_define_method(cTime, "mon", time_mon, 0);
+    rb_define_method(cTime, "month", time_mon, 0);
     rb_define_method(cTime, "year", time_year, 0);
     rb_define_method(cTime, "wday", time_wday, 0);
     rb_define_method(cTime, "yday", time_yday, 0);
@@ -792,4 +892,8 @@ Init_Time()
 #if defined(HAVE_TIMES) || defined(NT)
     S_Tms = struct_define("Tms", "utime", "stime", "cutime", "cstime", 0);
 #endif
+
+    /* methods for marshaling */
+    rb_define_singleton_method(cTime, "_load", time_load, 1);
+    rb_define_method(cTime, "_dump", time_dump, 1);
 }
